# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

__version__ = '2.22.4'

#: Chemical engineering plant cost index (defaults to 567.5 at 2017).
CE = 567.5 

#: Whether to add label the ID of streams with sources and sinks in process 
#: flow diagrams.
LABEL_PROCESS_STREAMS_IN_DIAGRAMS = True

#: Whether to automatically generate diagrams when displaying an object in the
#: IPython console.
ALWAYS_DISPLAY_DIAGRAMS = True

#: Whether to ignore unit graphics and display unit nodes as dots in process
#: flow diagrams.
MINIMAL_UNIT_DIAGRAMS = False

#: Whether to number unit operations in diagrams according to their order in the system path.
LABEL_PATH_NUMBER_IN_DIAGRAMS = False

# %% Initialize BioSTEAM 

from flexsolve import speed_up
from ._heat_utility import HeatUtility, UtilityAgent
from ._power_utility import PowerUtility
from ._unit import Unit
from ._system import System
from ._tea import CombinedTEA, TEA
from ._flowsheet import Flowsheet, main_flowsheet
from ._network import Network
from . import utils
from . import units
from . import evaluation
from . import exceptions
from . import process_tools
from . import examples
from . import report
from . import digraph

__all__ = ('Unit', 'PowerUtility', 'HeatUtility', 'UtilityAgent',
           'System', 'TEA', 'CombinedTEA', 'utils', 'units', 'evaluation', 
           'main_flowsheet', 'Flowsheet', 'CE', 'Chemical', 'Chemicals', 'Stream',
           'MultiStream', 'settings', 'Network', 'exceptions', 'speed_up', 'report',
           'digraph', 'process_tools', 'examples', *units.__all__, *evaluation.__all__
)

from thermosteam import Chemical, Chemicals, Thermo, Stream, MultiStream, settings
from .evaluation import *
from .units import *



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
General functional algorithms for the design of heat exchangers.

"""
from numpy import log as ln
from flexsolve import njitable

__all__ = ('counter_current_heat_exchange',
           'heuristic_overall_heat_transfer_coefficient',
           'heuristic_pressure_drop',
           'heuristic_tubeside_and_shellside_pressure_drops',
           'order_streams',
           'compute_Fahkeri_LMTD_correction_factor',
           'compute_heat_transfer_area',
           'compute_LMTD')

# %% Functional heat exchanger

def heat_exchange_to_condition(s_in, s_out, T, phase=None):
    """
    Set the outlet stream condition and return duty required to achieve 
    said condition.
    """
    if phase:
        s_out.T = T
        s_out.phase = phase
    else:
        s_out.vle(T=T, P=s_out.P)
    return s_out.H - s_in.H

def counter_current_heat_exchange(s0_in, s1_in, s0_out, s1_out,
                                  dT, T_lim0=None, T_lim1=None,
                                  phase0=None, phase1=None):
    """
    Allow outlet streams to exchange heat until either the given temperature 
    limits or the pinch temperature and return the total heat transfer 
    [Q; in kJ/hr].
    """
    # Counter_current_heat_exchange_setup:
    # First find the hot inlet, cold inlet, hot outlet and cold outlet streams
    # along with the maximum temperature approaches for the hotside and the 
    # cold side.
    if s0_in.T > s1_in.T:
        s_hot_in = s0_in
        s_cold_in = s1_in
        s_hot_out = s0_out
        s_cold_out = s1_out
        T_lim_coldside = T_lim0
        T_lim_hotside = T_lim1
        phase_coldside = phase0
        phase_hotside = phase1
    else:
        s_cold_in = s0_in
        s_hot_in = s1_in
        s_cold_out = s0_out
        s_hot_out = s1_out
        T_lim_hotside = T_lim0
        T_lim_coldside = T_lim1
        phase_hotside = phase0
        phase_coldside = phase1
    
    if (s_hot_in.T - s_cold_in.T) <= dT: return 0. # No heat exchange
    
    T_pinch_coldside = s_cold_in.T + dT
    if T_lim_coldside:
        T_lim_coldside = max(T_pinch_coldside, T_lim_coldside) 
    else:
        T_lim_coldside = T_pinch_coldside
    
    T_pinch_hotside = s_hot_in.T - dT
    if T_lim_hotside:
        T_lim_hotside = min(T_pinch_hotside, T_lim_hotside) 
    else:
        T_lim_hotside = T_pinch_hotside
       
    # Find which side reaches the pinch first by selecting the side that needs
    # the least heat transfer to reach the pinch.
       
    # Pinch on the cold side
    Q_hot_stream = heat_exchange_to_condition(s_hot_in, s_hot_out, 
                                              T_lim_coldside, phase_coldside)
    
    # Pinch on the hot side
    Q_cold_stream = heat_exchange_to_condition(s_cold_in, s_cold_out, 
                                               T_lim_hotside, phase_hotside)
    
    if Q_hot_stream > 0 or Q_cold_stream < 0:
        # Sanity check
        raise RuntimeError('inlet stream not in vapor-liquid equilibrium')
    
    if Q_cold_stream < -Q_hot_stream:
        # Pinch on the hot side            
        Q = Q_cold_stream
        if phase_coldside:
            s_hot_out.H = s_hot_in.H - Q
        else:
            s_hot_out.vle(H=s_hot_in.H - Q, P=s_hot_out.P)
    else:
        # Pinch on the cold side
        Q = Q_hot_stream
        if phase_hotside:
            s_cold_out.H = s_cold_in.H - Q
        else:
            s_cold_out.vle(H=s_cold_in.H - Q, P=s_cold_out.P)
    
    return abs(Q)


# %% Heuristics

def heuristic_overall_heat_transfer_coefficient(ci, hi, co, ho):
    """
    Return a heuristic estimate of the overall heat transfer coefficient
    [U; in kW/m^2/K]. Assume `U` is 0.5 kW/m^2/K if heat exchange is sensible 
    and 1.0 kW/m^2/K otherwise.
    
    Parameters
    ----------
    ci : Stream
        Cold inlet stream.
    hi : Stream
        Hot inlet stream.
    co : Stream
        Cold outlet stream.
    ho : Stream
        Hot outlet stream.
    
    Returns
    -------
    U : float
        overall heat transfer coefficient [kW/m^2/K].
    
    """
    # TODO: Base U on Table 18.5, Warren D. Seider et. al. Product and Process Design Principles. (2016)
    cip, hip, cop, hop = ci.phase, hi.phase, co.phase, ho.phase
    phases = cip + hip + cop + hop
    if 'g' in phases:
        if ('g' in hip and 'l' in hop) and ('l' in cip and 'g' in cop):
            return 1.0
        else:
            return 0.5
    else:
        return 0.5

def heuristic_pressure_drop(inlet_phase, outlet_phase):
    """
    Return a heuristic estimate of the pressure drop [dP; in psi]. If the fluid 
    changes phase, `dP` is 1.5 psi. If the fluid remains a liquid, `dP` is 5 psi.
    If the fluid remains a gas, `dP` is 3 psi.
    
    Parameters
    ----------
    inlet_phase: str
    outlet_phase : str
        
    Returns
    -------
    dP : float
        Pressure drop [psi].
    
    """
    if ('l' in inlet_phase and 'g' in outlet_phase) or ('g' in inlet_phase and 'l' in outlet_phase):
        # Latent fluid (boiling or condensing)
        dP = 1.5
    elif inlet_phase == 'l':
        # Sensible liquid
        dP = 5
    elif outlet_phase == 'g':
        # Sensible vapor
        dP = 3
    return dP

def heuristic_tubeside_and_shellside_pressure_drops(ci, hi, co, ho,
                                                    tubeside_iscooling=True):
    """
    Return an estimate of tubeside and shellside pressure drops.
    
    Parameters
    ----------
    ci : Stream
        Cold inlet stream.
    hi : Stream
        Hot inlet stream.
    co : Stream
        Cold outlet stream.
    ho : Stream
        Hot outlet stream.
    tubeside_iscooling : bool
        True of tubeside fluid is cooling.
    
    Returns
    -------  
    dP_tube : float
        Tubeside pressure drop (psi)
    dP_shell : float
        Shellside pressure drop (psi)
    
    """
    dP_c = heuristic_pressure_drop(ci.phase, co.phase)
    dP_h = heuristic_pressure_drop(hi.phase, ho.phase)
    if tubeside_iscooling:
        dP_tube = dP_h
        dP_shell = dP_c
    else:
        dP_tube = dP_c
        dP_shell = dP_h
    return dP_tube, dP_shell

# %% General functions

def order_streams(in_a, in_b, out_a, out_b):
    """
    Return cold and hot inlet and outlet streams.
    
    Parameters
    ----------
    in_a : Stream
        Inlet a.
    in_b : Stream
        Inlet b.
    out_a : Stream
        Outlet a.
    out_b : Stream
        Outlet b.
    
    Returns
    -------
    ci : Stream
        Cold inlet.
    hi : Stream
        Hot inlet.
    co : Stream
        Cold outlet.
    ho : Stream
        Hot outlet.
    
    """
    if in_a.T < in_b.T:
        return in_a, in_b, out_a, out_b
    else:
        return in_b, in_a, out_b, out_a

# %% Computational functions

@njitable(cache=True)
def compute_fallback_Fahkeri_LMTD_correction_factor(P, N_shells):
    """Return LMTF correction factor using the fallback equation for
    `compute_Fahkeri_LMTD_correction_factor` when logarithms cannot be computed."""
    # A, J, and K are dummy variables
    A = N_shells - N_shells*P
    W = A/(A + P)
    if 0.999 < W < 1.001:
        Ft = 1
    else:
        J = W/(1. - W)
        K = (J + 2**-0.5)/(J - 2**-0.5)
        if K <= 1:
            Ft = 1
        else:
            Ft = (2**0.5*J)/ln(K)
    return Ft

@njitable(cache=True)
def compute_Fahkeri_LMTD_correction_factor(Tci, Thi, Tco, Tho, N_shells):
    r"""
    Return the log-mean temperature difference correction factor `Ft` 
    for a shell-and-tube heat exchanger with one or an even number of tube 
    passes, and a given number of shell passes, with the expression given in 
    [1]_ and also shown in [2]_.
    
    .. math::
        F_t=\frac{S\ln W}{\ln \frac{1+W-S+SW}{1+W+S-SW}}
    
        S = \frac{\sqrt{R^2+1}}{R-1}
        
        W = \left(\frac{1-PR}{1-P}\right)^{1/N}
        
        R = \frac{T_{in}-T_{out}}{t_{out}-t_{in}}
        
        P = \frac{t_{out}-t_{in}}{T_{in}-t_{in}}
        
    If R = 1 and logarithms cannot be evaluated:
        
    .. math::
        W' = \frac{N-NP}{N-NP+P}
        
        F_t = \frac{\sqrt{2}\frac{1-W'}{W'}}{\ln\frac{\frac{W'}{1-W'}+\frac{1}
        {\sqrt{2}}}{\frac{W'}{1-W'}-\frac{1}{\sqrt{2}}}}
        
    Parameters
    ----------
    Tci : float
        Inlet temperature of cold fluid, [K]
    Thi : float
        Inlet temperature of hot fluid, [K]
    Tco : float
        Outlet temperature of cold fluid, [K]        
    Tho : float
        Outlet temperature of hot fluid, [K]
    shells : int, optional
        Number of shell-side passes, [-]
    
    Returns
    -------
    Ft : float
        Log-mean temperature difference correction factor, [-]
    
    Notes
    -----
    This expression is symmetric - the same result is calculated if the cold
    side values are swapped with the hot side values. It also does not 
    depend on the units of the temperature given.
    
    Examples
    --------
    compute_Fahkeri_LMTD_correction_factor(Tci=15, Tco=85, Thi=130, Tho=110, N_shells=1)
    0.9438358829645933
    
    References
    ----------
    .. [1] Fakheri, Ahmad. "A General Expression for the Determination of the 
       Log Mean Temperature Correction Factor for Shell and Tube Heat 
       Exchangers." Journal of Heat Transfer 125, no. 3 (May 20, 2003): 527-30.
       doi:10.1115/1.1571078.
    .. [2] Hall, Stephen. Rules of Thumb for Chemical Engineers, Fifth Edition.
       Oxford; Waltham, MA: Butterworth-Heinemann, 2012.
    
    """
    if (Tco - Tci) < 0.01:
        R = 1
    else:
        R = (Thi - Tho)/(Tco - Tci)
    P = (Tco - Tci)/(Thi - Tci)
    if 0.999 < R < 1.001:
        Ft = compute_fallback_Fahkeri_LMTD_correction_factor(P, N_shells)
    else:
        W = ((1. - P*R)/(1. - P))**(1./N_shells)
        S = (R*R + 1.)**0.5/(R - 1.)
        K = (1. + W - S + S*W)/(1. + W + S - S*W)
        if K <= 0.001 or 0.999 < K < 1.001:
            Ft = compute_fallback_Fahkeri_LMTD_correction_factor(P, N_shells)
        else:
            Ft = S*ln(W)/ln(K)
    if Ft > 1.0:
        Ft = 1.0
    elif Ft < 0.5:
        # Bad design, probably a heat exchanger network operating
        # too close to the pinch. Fahkeri may not be valid, so give
        # a conservative estimate of the correction factor.
        Ft = 0.5
    return Ft

@njitable(cache=True)
def compute_heat_transfer_area(LMTD, U, Q, ft):
    """
    Return required heat transfer area by LMTD correction factor method.
    
    Parameters
    ----------
    LMTD : float
        Log mean temperature difference
    U : float
        Heat transfer coefficient
    Q : float
        Duty
    
    """
    return Q/(U*LMTD*ft)   

@njitable(cache=True)
def compute_LMTD(Thi, Tho, Tci, Tco, counterflow=True):
    r'''
    Return the log-mean temperature difference of an ideal counterflow
    or co-current heat exchanger.

    .. math::
        \Delta T_{LMTD}=\frac{\Delta T_1-\Delta T_2}{\ln(\Delta T_1/\Delta T_2)}

        \text{For countercurrent:      } \\
        \Delta T_1=T_{h,i}-T_{c,o}\\
        \Delta T_2=T_{h,o}-T_{c,i}

        \text{Parallel Flow Only:} \\
        {\Delta T_1=T_{h,i}-T_{c,i}}\\
        {\Delta T_2=T_{h,o}-T_{c,o}}

    Parameters
    ----------
    Thi : float
        Inlet temperature of hot fluid [K].
    Tho : float
        Outlet temperature of hot fluid [K].
    Tci : float
        Inlet temperature of cold fluid [K].
    Tco : float
        Outlet temperature of cold fluid [K].
    counterflow : bool, optional
        Whether the exchanger is counterflow or co-current.

    Returns
    -------
    LMTD : float
        Log-mean temperature difference [K]

    Notes
    -----
    Any consistent set of units produces a consistent output.

    Examples
    --------
    >>> compute_LMTD(100., 60., 30., 40.2)
    43.200409294131525
    >>> compute_LMTD(100., 60., 30., 40.2, counterflow=False)
    39.75251118049003
    
    '''
    if counterflow:
        dTF1 = Thi-Tco
        dTF2 = Tho-Tci
    else:
        dTF1 = Thi-Tci
        dTF2 = Tho-Tco
    dTF21 = dTF2 - dTF1
    if abs(dTF21) < 1e-8:
        LMTD = dTF1
    else:
        LMTD = dTF21/ln(dTF2/dTF1)
    return LMTD


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .. import Unit, units
from ..utils import format_unit_name
from . import decorators
import pandas as pd
import numpy as np

_add_cost = decorators.add_cost
_index = np.array(('Basis',
                   'Units',
                   'Size',
                   'Upper bound',
                   'CEPCI',
                   'Cost (USD)',
                   'Exponent',
                   'Electricity (kW)',
                   'Installation factor',
                   'Number'))


__all__ = ('df2unit', 'xl2dct', 'xl2mod')

def df2unit(clsname, cost_items, *, supercls=None, metacls=None):
    """Return Unit subclass from cost_options DataFrame."""
    superclasses = (supercls,) if supercls else (Unit,)
    if not metacls: metacls = type(supercls)
    cls = metacls.__new__(metacls, clsname, superclasses, {})
    for ID in cost_items:
        _add_cost(cls, ID, *cost_items[ID], None)
    return cls

def df2dct(df):
    dct = {}
    for name_sim in df.columns.levels[0]:
        if name_sim == 'Unit': continue
        cost_items = df[name_sim]
        if '-' in name_sim:
            sim, name = name_sim.split('-')
            is_static = False
        else:
            name = name_sim
            sim = 'Unit'
            is_static = True
        name = format_unit_name(name)
        try:
            if is_static:
                supercls = Unit
            else:
                supercls = getattr(units, sim)
        except:
            supername = ''.join([i.capitalize() for i in sim.split(' ')])
            try: supercls = getattr(units, supername)
            except AttributeError:
                raise ValueError(f"invalid simulation option '{sim}'")
        dct[name] = df2unit(name, cost_items, supercls=supercls)
    return dct

def xl2dct(file, sheet_name=0):
    """Return dictionary of unit subclasses from excel file."""
    return df2dct(pd.read_excel(file, header=[0, 1], nrows=10))

def xl2mod(file, module, sheet_name=0):
    dct = xl2dct(file, sheet_name)
    for i, j in dct.items():
        setattr(module, i, j)
        j.__module__ = module.__name__
    if not hasattr(module, '__all__'):
        module.__all__ = tuple(dct)
    
    
    
    
    
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import digraph

__all__ = (*digraph.__all__,)

from .digraph import *


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
This module contains heat exchanger unit operations.

.. contents:: :local:
    
Unit Operations
---------------
.. autoclass:: biosteam.units.heat_exchange.HX
.. autoclass:: biosteam.units.heat_exchange.HXutility
.. autoclass:: biosteam.units.heat_exchange.HXprocess 

"""
from .. import Unit
from .._graphics import utility_heat_exchanger_graphics, process_heat_exchanger_graphics
from .design_tools.specification_factors import (
    shell_and_tube_material_factor_coefficients,
    compute_shell_and_tube_material_factor)
from .design_tools import heat_transfer as ht
import numpy as np
import biosteam as bst
from math import exp, log as ln

__all__ = ('HX', 'HXutility', 'HXprocess')

# Lenght factor 
x = np.array((8, 13, 16, 20)) 
y = np.array((1.25, 1.12,1.05,1))
p2 = np.polyfit(x, y, 2)

# %% Purchase price

def compute_floating_head_purchase_price(A, CE):
    return exp(12.0310 - 0.8709*ln(A) + 0.09005 * ln(A)**2)*CE/567

def compute_fixed_head_purchase_price(A, CE):
    return exp(11.4185 - 0.9228*ln(A) + 0.09861 * ln(A)**2)*CE/567

def compute_u_tube_purchase_price(A, CE):
    return exp(11.5510 - 0.9186*ln(A) + 0.09790 * ln(A)**2)*CE/567

def compute_kettle_vaporizer_purchase_price(A, CE):
    return exp(12.3310 - 0.8709*ln(A) + 0.09005 * ln(A)**2)*CE/567

def compute_double_pipe_purchase_price(A, CE):
    return exp( 7.2718 + 0.16*ln(A))*CE/567

# Purchase price
Cb_dict = {'Floating head': compute_floating_head_purchase_price,
           'Fixed head': compute_fixed_head_purchase_price,
           'U tube': compute_u_tube_purchase_price,
           'Kettle vaporizer': compute_kettle_vaporizer_purchase_price,
           'Double pipe': compute_double_pipe_purchase_price}

# %% Classes

class HX(Unit, isabstract=True):
    """
    Abstract class for counter current heat exchanger.

    **Abstract methods**
    
    get_streams()
        Should return two inlet streams and two outlet streams that exchange
        heat.

    """
    line = 'Heat Exchanger'
    _units = {'Area': 'ft^2',
              'Overall heat transfer coefficient': 'kW/m^2/K',
              'Log-mean temperature difference': 'K',
              'Tube side pressure drop': 'psi',
              'Shell side pressure drop': 'psi',
              'Operating pressure': 'psi',
              'Total tube length': 'ft'}
    _N_ins = 1
    _N_outs = 1
    _N_heat_utilities = 1
    _BM = {'Double pipe': 1.8,
           'Floating head': 3.17,
           'Fixed head': 3.17,
           'U tube': 3.17,
           'Kettle vaporizer': 3.17}
    
    @property
    def material(self):
        """Default 'Carbon steel/carbon steel'"""
        return self.material
    @material.setter
    def material(self, material):
        try:
            self._F_Mab = shell_and_tube_material_factor_coefficients[material]
        except KeyError:
            raise AttributeError("material must be one of the following: "
                                 f"{', '.join(shell_and_tube_material_factor_coefficients)}")
        self._material = material  
    
    @property
    def heat_exchanger_type(self):
        """[str] Heat exchanger type. Purchase cost depends on this selection."""
        return self._heat_exchanger_type
    @heat_exchanger_type.setter
    def heat_exchanger_type(self, heat_exchanger_type):
        try:
            self._Cb_func = Cb_dict[heat_exchanger_type]
        except KeyError:
            raise AttributeError("heat exchange type must be one of the following: "
                                 f"{', '.join(Cb_dict)}")
        self._heat_exchanger_type = heat_exchanger_type     

    def _design(self):
        # Get duty (kW)
        Q = abs(self.Q) / 3600
        
        if Q <= 1e-12: 
            self.design_results.clear()
            return
        
        ###  Use LMTD correction factor method  ###
        Design = self.design_results
        
        # Get cold and hot inlet and outlet streams
        ci, hi, co, ho = ht.order_streams(*self.get_streams())
       
        # Get log mean temperature difference
        Tci = ci.T
        Thi = hi.T
        Tco = co.T
        Tho = ho.T
        LMTD = ht.compute_LMTD(Thi, Tho, Tci, Tco)
        
        # Get correction factor
        ft = self.ft
        if not ft:
            N_shells = self.N_shells
            ft = ht.compute_Fahkeri_LMTD_correction_factor(Tci, Thi, Tco, Tho, N_shells)
        
        # Get overall heat transfer coefficient
        U = self.U or ht.heuristic_overall_heat_transfer_coefficient(ci, hi, co, ho)
        dP_tube, dP_shell = ht.heuristic_tubeside_and_shellside_pressure_drops(ci, hi, co, ho)
        
        # TODO: Complete design of heat exchanger to find L
        # For now assume lenght is 20 ft
        L = 20
        
        # Design pressure
        P = max((ci.P, hi.P))
        Design['Area'] = 10.763 * ht.compute_heat_transfer_area(abs(LMTD), U, Q, ft)
        Design['Overall heat transfer coefficient'] = U
        Design['Log-mean temperature difference'] = LMTD
        Design['Fouling correction factor'] = ft
        Design['Tube side pressure drop'] = dP_tube
        Design['Shell side pressure drop'] = dP_shell
        Design['Operating pressure'] = P * 14.7/101325 # psi
        Design['Total tube length'] = L

    def _cost(self):
        Design = self.design_results
        if not Design: 
            self.purchase_costs.clear()
            return
            
        A = Design['Area']
        L = Design['Total tube length']
        P = Design['Operating pressure']
        
        if A < 150: # Double pipe
            P = P/600
            F_p = 0.8510 + 0.1292*P + 0.0198*P**2
            # Assume outer pipe carbon steel, inner pipe stainless steel
            F_m = 2 
            A_min = 2.1718
            if A < A_min:
                F_l = A/A_min
                A = A_min
            else:    
                F_l = 1
            heat_exchanger_type = 'Double pipe'
            C_b = compute_double_pipe_purchase_price(A, bst.CE)
        else: # Shell and tube
            F_m = compute_shell_and_tube_material_factor(A,  *self._F_Mab)
            F_l = 1 if L > 20 else np.polyval(p2, L)
            P = P/100
            F_p = 0.9803 + 0.018*P + 0.0017*P**2
            heat_exchanger_type = self.heat_exchanger_type
            C_b = self._Cb_func(A, bst.CE)
        
        # Free on board purchase prize
        self.purchase_costs[heat_exchanger_type] = F_p * F_l * F_m * C_b


class HXutility(HX):
    """
    Create a heat exchanger that changes temperature of the
    outlet stream using a heat utility.

    Parameters
    ----------
    ins : stream
        Inlet.
    outs : stream
        Outlet.
    T=None : float
        Temperature of outlet stream [K].
    V=None : float
        Vapor fraction of outlet stream.
    rigorous=False : bool
        If true, calculate vapor liquid equilibrium
    U=None : float, optional
        Enforced overall heat transfer coefficent [kW/m^2/K].
    heat_exchanger_type : str, optional
        Heat exchanger type. Defaults to "Floating head".
    N_shells=2 : int
        Number of shells.
    ft=None : float
        User imposed correction factor.
    
    Notes
    -----
    Must specify either `T` or `V` when creating a HXutility object.
    
    Examples
    --------
    Run heat exchanger by temperature:
    
    >>> from biosteam.units import HXutility
    >>> from biosteam import Stream, settings
    >>> settings.set_thermo(['Water', 'Ethanol'], cache=True)
    >>> feed = Stream('feed', Water=200, Ethanol=200)
    >>> hx = HXutility('hx', ins=feed, outs='product', T=50+273.15,
    ...                rigorous=False) # Ignore VLE
    >>> hx.simulate()
    >>> hx.show()
    HXutility: hx
    ins...
    [0] feed
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water    200
                        Ethanol  200
    outs...
    [0] product
        phase: 'l', T: 323.15 K, P: 101325 Pa
        flow (kmol/hr): Water    200
                        Ethanol  200
    >>> hx.results()
    Heat Exchanger                                            Units       hx
    Low pressure steam  Duty                                  kJ/hr 1.01e+06
                        Flow                                kmol/hr     26.1
                        Cost                                 USD/hr     6.21
    Design              Area                                   ft^2       57
                        Overall heat transfer coefficient  kW/m^2/K      0.5
                        Log-mean temperature difference           K      101
                        Fouling correction factor                          1
                        Tube side pressure drop                 psi      1.5
                        Shell side pressure drop                psi        5
                        Operating pressure                      psi       50
                        Total tube length                        ft       20
    Purchase cost       Double pipe                             USD 4.74e+03
    Total purchase cost                                         USD 4.74e+03
    Utility cost                                             USD/hr     6.21
    
    Run heat exchanger by vapor fraction:
    
    >>> feed = Stream('feed', Water=200, Ethanol=200)
    >>> hx = HXutility('hx', ins=feed, outs='product', V=1,
    ...                rigorous=True) # Include VLE
    >>> hx.simulate()
    >>> hx.show()
    HXutility: hx
    ins...
    [0] feed
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water    200
                        Ethanol  200
    outs...
    [0] product
        phase: 'g', T: 357.45 K, P: 101325 Pa
        flow (kmol/hr): Water    200
                        Ethanol  200
    >>> hx.results()
    Heat Exchanger                                            Units       hx
    Low pressure steam  Duty                                  kJ/hr 1.94e+07
                        Flow                                kmol/hr      499
                        Cost                                 USD/hr      119
    Design              Area                                   ft^2      680
                        Overall heat transfer coefficient  kW/m^2/K        1
                        Log-mean temperature difference           K     80.8
                        Fouling correction factor                          1
                        Tube side pressure drop                 psi      1.5
                        Shell side pressure drop                psi      1.5
                        Operating pressure                      psi       50
                        Total tube length                        ft       20
    Purchase cost       Floating head                           USD 2.61e+04
    Total purchase cost                                         USD 2.61e+04
    Utility cost                                             USD/hr      119
    
    """
    line = 'Heat Exchanger'
    _graphics = utility_heat_exchanger_graphics
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *,
                 T=None, V=None, rigorous=False, U=None,
                 heat_exchanger_type="Floating head",
                 material="Carbon steel/carbon steel",
                 N_shells=2,
                 ft=None):
        super().__init__(ID, ins, outs, thermo)
        self.T = T #: Temperature of outlet stream (K).
        self.V = V #: Vapor fraction of outlet stream.
        
        #: [bool] If true, calculate vapor liquid equilibrium
        self.rigorous = rigorous
        
        #: [float] Enforced overall heat transfer coefficent (kW/m^2/K)
        self.U = U
        
        #: [float] Total heat transfered.
        self.Q = None
        
        #: Number of shells for LMTD correction factor method.
        self.N_shells = N_shells
        
        #: User imposed correction factor.
        self.ft = ft
        
        self.material = material
        self.heat_exchanger_type = heat_exchanger_type
    
    def _init_utils(self):
        # tuple[HeatUtility] All heat utilities associated to unit
        self._heat_utilities = tuple([bst.HeatUtility(heat_exchanger=self)
                                      for i in range(self._N_heat_utilities)])
        
        # [PowerUtility] Electric utility associated to unit
        self.power_utility = bst.PowerUtility()
    
    @property
    def heat_utilities(self):
        return self._heat_utilities
    @heat_utilities.setter
    def heat_utilities(self, heat_utilities):
        self._heat_utilities = heat_utilities = tuple(heat_utilities)
        for i in heat_utilities: i.heat_exchanger = self
    
    def simulate_as_auxiliary_exchanger(self, duty, stream):
        self.outs[0] = stream.proxy()
        self.ins[0] = stream.proxy()
        hu = self.heat_utilities[0]
        hu.heat_exchanger = None
        hu(duty, stream.T)
        self.Q = duty
        super()._design()
        self._cost()
        
    def _run(self):
        feed = self.ins[0]
        s = self.outs[0]
        s.copy_flow(feed)
        s.P = feed.P
        T = self.T
        V = self.V
        V_given = V is not None
        if self.rigorous:
            if T and V_given:
                raise RuntimeError("may only define either temperature, 'T', or "
                                   "vapor fraction 'V', in a rigorous simulation")
            if V_given:
                if V == 0:
                    s.phase = 'l'
                    s.T = s.bubble_point_at_P().T
                elif V == 1:
                    s.phase = 'g'
                    s.T = s.dew_point_at_P().T
                elif 0 < V < 1:
                    s.vle(V=V, P=s.P)
                else:
                    raise RuntimeError("vapor fraction, 'V', must be a "
                                       "positive fraction")
            else:
                s.vle(T=T, P=s.P)
        elif (T or V_given):
            if V_given:
                if V == 0:
                    s.phase = 'l'
                elif V == 1:
                    s.phase = 'g'
                else:
                    raise RuntimeError("vapor fraction, 'V', must be either "
                                       "0 or 1 in a non-rigorous simulation")
            else:
                phase = feed.phase
                if len(phase) == 1: s.phase = phase
            if T:
                s.T = T
            else:
                s.T = feed.T
        else:
            raise RuntimeError("must define either temperature, 'T', and/or "
                               "vapor fraction, 'V', in a non-rigorous "
                               "simulation")

    def get_streams(self):
        """
        Return inlet and outlet streams.
        
        Returns
        -------
        in_a : Stream
            Inlet a.
        in_b : Stream
            Inlet b.
        out_a : Stream
            Outlet a.
        out_b : Stream
            Outlet b.
        
        """
        in_a = self.ins[0]
        out_a = self.outs[0]
        hu = self.heat_utilities[0]
        in_b = hu.inlet_utility_stream
        out_b = hu.outlet_utility_stream
        return in_a, in_b, out_a, out_b

    def _design(self, duty=None):
        # Set duty and run heat utility
        if duty is None:
            duty = self.H_out - self.H_in
        self.heat_utilities[0](duty, self.ins[0].T, self.outs[0].T)
        self.Q = duty
        super()._design()


class HXprocess(HX):
    """
    Counter current heat exchanger for process fluids. Rigorously transfers
    heat until the pinch temperature or a user set temperature limit is reached.
    
    Parameters
    ----------
    ins : stream sequence
        * [0] Inlet process fluid a
        * [1] Inlet process fluid b  
    outs : stream sequence
        * [0] Outlet process fluid a        
        * [1] Outlet process fluid b
    U=None : float, optional
        Enforced overall heat transfer coefficent [kW/m^2/K].
    dT=5. : float
        Pinch temperature difference (i.e. dT = abs(outs[0].T - outs[1].T)).
    T_lim0 : float, optional
        Temperature limit of outlet stream at index 0.
    T_lim1 : float, optional
        Temperature limit of outlet stream at index 1.
    heat_exchanger_type : str, optional
        Heat exchanger type. Defaults to 'Floating head'.
    N_shells=2 : int, optional
        Number of shells.
    ft=None : float, optional
        User enforced correction factor.
    phase0=None : 'l' or 'g', optional
        User enforced phase of outlet stream at index 0.
    phase1=None : 'l' or 'g', optional
        User enforced phase of outlet stream at index 1.
    
    Examples
    --------
    Rigorous heat exchange until pinch temperature is reached:
        
    >>> from biosteam.units import HXprocess
    >>> from biosteam import Stream, settings
    >>> settings.set_thermo(['Water', 'Ethanol'])
    >>> in_a = Stream('in_a', Ethanol=50, T=351.43, phase='g')
    >>> in_b = Stream('in_b', Water=200)
    >>> hx = HXprocess('hx', ins=(in_a, in_b), outs=('out_a', 'out_b'))
    >>> hx.simulate()
    >>> hx.show()
    HXprocess: hx
    ins...
    [0] in_a
        phase: 'g', T: 351.43 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  50
    [1] in_b
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Water  200
    outs...
    [0] out_a
        phases: ('g', 'l'), T: 351.39 K, P: 101325 Pa
        flow (kmol/hr): (g) Ethanol  31.3
                        (l) Ethanol  18.7
    [1] out_b
        phases: ('g', 'l'), T: 346.43 K, P: 101325 Pa
        flow (kmol/hr): (l) Water  200
    
    >>> hx.results()
    Heat Exchanger                                            Units       hx
    Design              Area                                   ft^2      213
                        Overall heat transfer coefficient  kW/m^2/K      0.5
                        Log-mean temperature difference           K     20.4
                        Fouling correction factor                          1
                        Tube side pressure drop                 psi      1.5
                        Shell side pressure drop                psi        5
                        Operating pressure                      psi     14.7
                        Total tube length                        ft       20
    Purchase cost       Floating head                           USD 2.06e+04
    Total purchase cost                                         USD 2.06e+04
    Utility cost                                             USD/hr        0
    
    Sensible fluids case with user enfored outlet phases 
    (more computationally efficient):
        
    >>> from biosteam.units import HXprocess
    >>> from biosteam import Stream, settings
    >>> settings.set_thermo(['Water', 'Ethanol'])
    >>> in_a = Stream('in_a', Water=200, T=350)
    >>> in_b = Stream('in_b', Ethanol=200)
    >>> hx = HXprocess('hx', ins=(in_a, in_b), outs=('out_a', 'out_b'),
    ...                phase0='l', phase1='l')
    >>> hx.simulate()
    >>> hx.show()
    HXprocess: hx
    ins...
    [0] in_a
        phase: 'l', T: 350 K, P: 101325 Pa
        flow (kmol/hr): Water  200
    [1] in_b
        phase: 'l', T: 298.15 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  200
    outs...
    [0] out_a
        phase: 'l', T: 303.15 K, P: 101325 Pa
        flow (kmol/hr): Water  200
    [1] out_b
        phase: 'l', T: 327.92 K, P: 101325 Pa
        flow (kmol/hr): Ethanol  200
    >>> hx.results()
    Heat Exchanger                                            Units       hx
    Design              Area                                   ft^2      367
                        Overall heat transfer coefficient  kW/m^2/K      0.5
                        Log-mean temperature difference           K     11.5
                        Fouling correction factor                          1
                        Tube side pressure drop                 psi        5
                        Shell side pressure drop                psi        5
                        Operating pressure                      psi     14.7
                        Total tube length                        ft       20
    Purchase cost       Floating head                           USD 2.23e+04
    Total purchase cost                                         USD 2.23e+04
    Utility cost                                             USD/hr        0
    
    """
    line = 'Heat Exchanger'
    _graphics = process_heat_exchanger_graphics
    _N_heat_utilities = 0
    _N_ins = 2
    _N_outs = 2
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, *,
                 U=None, dT=5., T_lim0=None, T_lim1=None,
                 material="Carbon steel/carbon steel",
                 heat_exchanger_type="Floating head",
                 N_shells=2, ft=None, 
                 phase0=None,
                 phase1=None):
        super().__init__(ID, ins, outs, thermo)
        
        #: [float] Enforced overall heat transfer coefficent (kW/m^2/K)
        self.U = U
        
        #: [float] Total heat transfered.
        self.Q = None
        
        #: Number of shells for LMTD correction factor method.
        self.N_shells = N_shells
        
        #: User imposed correction factor.
        self.ft = ft
        
        #: [float] Pinch temperature difference.
        self.dT = dT 
        
        #: [float] Temperature limit of outlet stream at index 0.
        self.T_lim0 = T_lim0
        
        #: [float] Temperature limit of outlet stream at index 1.
        self.T_lim1 = T_lim1
        
        #: Enforced phase of outlet at index 0
        self.phase0 = phase0
        
        #: Enforced phase of outlet at index 1
        self.phase1 = phase1
        
        self.material = material
        self.heat_exchanger_type = heat_exchanger_type
        
    def get_streams(self):
        s_in_a, s_in_b = self.ins
        s_out_a, s_out_b  = self.outs
        return s_in_a, s_in_b, s_out_a, s_out_b
    
    def _setup(self):
        for i in self._ins:
            if i.source: i.empty()
            
    def simulate(self):
        self._run()
        self._summary()
    
    def _run(self):
        s1_in, s2_in = self._ins
        s1_out, s2_out = self._outs
        if s1_in.isempty():
            s1_out.empty()
            s2_out.copy_like(s2_in)
            self.Q = 0.
        elif s2_in.isempty():
            s2_out.empty()
            s1_out.copy_like(s1_in)
            self.Q = 0.
        else:
            s1_out.copy_like(s1_in)
            s2_out.copy_like(s2_in)
            self._run_counter_current_heat_exchange()
    
    def _run_counter_current_heat_exchange(self):
        self.Q = ht.counter_current_heat_exchange(*self._ins, *self._outs,
                                                  self.dT, self.T_lim0, self.T_lim1,
                                                  self.phase0, self.phase1)



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import Facility
from ..decorators import cost
from ... import HeatUtility
# from copy import copy

__all__ = ('CoolingTower',) #'CoolingTowerWithPowerDemand')


# %%

@cost('Flow rate', 'Cooling water pump',
      S=609624, kW=1021, cost=283671, CE=551, n=0.8, BM=3.1)
@cost('Flow rate', 'Cooling tower',
      S=557183, kW=1598, cost=1375e3, CE=551, n=0.7, BM=1.5)
class CoolingTower(Facility):
    """
    Create a cooling tower with capital cost and power based on the flow rate 
    of cooling water as in [1]_.
    
    Parameters
    ----------
    ID : str, optional
        Unit ID.
    
    References
    ----------
    .. [1] Humbird, D., Davis, R., Tao, L., Kinchin, C., Hsu, D., Aden, A.,
        Dudgeon, D. (2011). Process Design and Economics for Biochemical 
        Conversion of Lignocellulosic Biomass to Ethanol: Dilute-Acid 
        Pretreatment and Enzymatic Hydrolysis of Corn Stover
        (No. NREL/TP-5100-47764, 1013269). https://doi.org/10.2172/1013269
    
    """
    network_priority = 1
    _units = {'Flow rate': 'kmol/hr'}
    _N_heat_utilities = 1
    _N_outs = _N_ins = 2
    evaporation = 0.01
    blowdown = 0.001
    def __init__(self, ID='', agent=None):
        self.agent = cooling_water = agent or HeatUtility.get_cooling_agent('cooling_water')
        self.makeup_water = makeup_water = cooling_water.to_stream('cooling_tower_makeup_water')
        return_cooling_water = cooling_water.to_stream()
        cooling_water = return_cooling_water.flow_proxy()
        loss = makeup_water.flow_proxy()
        loss.ID = 'evaporation_and_blowdown'
        super().__init__(ID, (return_cooling_water, makeup_water),
                         (cooling_water, loss), thermo=cooling_water.thermo)
        self.cooling_water_utilities = set()
        
    def _run(self): pass
        
    def _load_utility_agents(self):
        cwu = self.cooling_water_utilities
        agent = self.agent
        cwu.clear()
        for u in self.system.units:
            if u is self: continue
            for hu in u.heat_utilities:
                if hu.agent is agent:
                    cwu.add(hu)
        
    def _design(self):
        cwu = self.cooling_water_utilities
        if not cwu: self._load_utility_agents()
        used = self._ins[0]
        hu = self.heat_utilities[0]
        self._load_utility_agents()
        hu.mix_from(cwu)
            
        used.imol['7732-18-5'] = \
        self.design_results['Flow rate'] = \
        self.cooling_water = hu.flow 
        cooling_water, loss = self.outs
        cooling_water.T = hu.inlet_utility_stream.T
        self.makeup_water.mol[0] = self.cooling_water * (self.evaporation + self.blowdown)
        loss.T = hu.outlet_utility_stream.T
        hu.reverse()




# -*- coding: utf-8 -*-
"""
Created on Sat May  2 16:44:24 2020

@author: sarangbhagwat
"""
import numpy as np
import copy
import biosteam as bst
from warnings import warn

__all__ = ('StreamLifeCycle', 'synthesize_network')

class LifeStage:
        
        def __init__(self, s_in, s_out):
            self.s_in = s_in
            self.s_out = s_out
        
        @property
        def H_in(self): return self.s_in.H
        
        @property
        def H_out(self): return self.s_out.H
        
        @property
        def unit(self): return self.s_in.sink
        
        def _info(self, N_tabs=1):
            tabs = N_tabs*'\t'
            return (f"{type(self).__name__}: {self.unit.ID}\n"
                    + tabs + f"H_in = {self.H_in:.3g} kJ\n"
                    + tabs + f"H_out = {self.H_out:.3g} kJ")
                    
        def __repr__(self):
            return (f"<{type(self).__name__}: {repr(self.unit)}, H_in = {self.H_in:.3g} kJ, H_out = {self.H_out:.3g} kJ>")
            
        def show(self):
            print(self._info())
        _ipython_display_ = show
        
        
class StreamLifeCycle:
    
    def __init__(self, index, cold):
        self.index = index
        self.name = 's_%s'%index
        self.cold = cold
        self.life_cycle = None
        
    def get_relevant_units(self, index, new_HXs, new_HX_utils):
        new_HXs_relevant = [hx for hx in new_HXs if '_%s_'%index in hx.ID]
        new_HX_utils_relevant = [hx for hx in new_HX_utils if '_%s_'%index in hx.ID]
        return new_HXs_relevant, new_HX_utils_relevant
        
    def get_life_cycle(self, new_HXs, new_HX_utils):
        index = self.index
        name = self.name
        cold = self.cold
        us_index = '_%s_'%index
        new_HXs_relevant, new_HX_utils_relevant =\
            self.get_relevant_units(index, new_HXs, new_HX_utils)
        life_cycle = [LifeStage(unit.ins[0], unit.outs[0])\
                      for unit in new_HXs_relevant if name in unit.ins[0].ID]\
                      + [LifeStage(unit.ins[1], unit.outs[1])\
                      for unit in new_HXs_relevant if name in unit.ins[1].ID]\
                      + [LifeStage(unit.ins[0], unit.outs[0])\
                      for unit in new_HX_utils_relevant if name in unit.ins[0].ID]
        life_cycle.sort(key = lambda pt: pt.H_out, reverse = not cold)
        self.life_cycle = life_cycle
        return life_cycle
        
    def __repr__(self):
        life_cycle = self.life_cycle
        cold = self.cold
        if not self.life_cycle:
            return 'Not initialized; run StreamLifeCycle.get_life_cycle or\
                  HX_Network.get_stream_life_cycles first.' 
        else:
            index = self.index
            name = 'Stream_%s'%index
            strtype = 'cold' if cold else 'hot'
            rep = ''
            for LifeStage in life_cycle:
                line = '\t\t' + repr(LifeStage) + '\n'
                rep += line
            rep = '<StreamLifeCycle: ' + name + ', ' + strtype  + '\n\tlife_cycle = [\n' +  rep[:-1] + '\n\t]>'
            return rep
        
    def show(self):
        info = repr(self).replace('[', '').replace(']', '').replace('life_cycle =', 'life_cycle:')
        print(info[1:-1])
        
    _ipython_display_ = show
    
    
def temperature_interval_pinch_analysis(hus, T_min_app = 10, ID_original = None):
    hx_utils = [hu for hu in hus\
                if abs(hu.heat_exchanger.ins[0].T - hu.heat_exchanger.outs[0].T)>0.01]
    hus_heating = [hu for hu in hx_utils if hu.duty > 0]
    hus_cooling = [hu for hu in hx_utils if hu.duty < 0]
    hxs_heating = [hu.heat_exchanger for hu in hx_utils if hu.duty > 0]
    hxs_cooling = [hu.heat_exchanger for hu in hx_utils if hu.duty < 0]
    streams_heating = [hx.ins[0].copy() for hx in hxs_heating]
    streams_cooling = [hx.ins[0].copy() for hx in hxs_cooling]
    init_heat_util, init_cool_util, init_Q = 0, 0, 0
    hx_utils_rearranged = hus_heating + hus_cooling
    hxs = hxs_heating + hxs_cooling
    streams = streams_heating + streams_cooling
    for i in range(len(streams)):
        stream = streams[i]
        hx = hxs[i]
        stream.vle(H = stream.H, P = stream.P)
        ID = 'Util_%s'%i
        stream.ID = 's_%s__%s'%(i,ID)
        
    len_hxs_heating = len(hxs_heating)
    is_cold_stream_index = lambda x: x<len_hxs_heating
    T_in_arr = np.array([stream.T for stream in streams])
    T_out_arr = np.array([hx.outs[0].T for hx in hxs])
    T_hot_side_arr = np.array([stream.T for stream in streams_heating]\
                               + [hx.outs[0].T for hx in hxs_cooling])
    T_cold_side_arr = np.array([hx.outs[0].T for hx in hxs_heating]\
                                + [stream.T for stream in streams_cooling])
    adj_T_in_arr = T_in_arr.copy()
    adj_T_in_arr[:len(hxs_heating)] -= T_min_app
    adj_T_out_arr = T_out_arr.copy()
    adj_T_out_arr[:len(hxs_heating)] -= T_min_app
    T_changes_tuples = list(zip(adj_T_in_arr, adj_T_out_arr))
    all_Ts_descending = [*adj_T_in_arr, *adj_T_out_arr]
    all_Ts_descending.sort(reverse=True)
    stream_indices_for_T_intervals =\
        {(all_Ts_descending[i], all_Ts_descending[i+1]):[]\
            for i in range(len(all_Ts_descending)-1)}
    H_for_T_intervals = dict.fromkeys(stream_indices_for_T_intervals, 0)
    cold_indices = list(range(len_hxs_heating))
    hot_indices = list(range(len_hxs_heating, len(hxs)))
    
    for i in range(len(all_Ts_descending)-1):
        T_start = all_Ts_descending[i]
        T_end = all_Ts_descending[i+1]
        dT = T_start - T_end
        for stream_index in range(len(T_changes_tuples)):
            if ((T_changes_tuples[stream_index][0]>= T_start and
                    T_changes_tuples[stream_index][1]<= T_end) or
                    (T_changes_tuples[stream_index][1]>= T_start and
                     T_changes_tuples[stream_index][0]<= T_end)):
                stream_indices_for_T_intervals[(T_start, T_end)].append(stream_index)
                multiplier = 1
                if is_cold_stream_index(stream_index):
                    multiplier = -1
                stream = streams[stream_index].copy()
                stream.vle(T = T_start, P = stream.P)
                H1 = stream.H
                stream.vle(T = T_end, P = stream.P)
                H2 = stream.H
                H = multiplier*(H1 - H2)
                H_for_T_intervals[(T_start, T_end)] += H
                
    res_H_vector = []
    prev_res_H = 0
    for interval, H in H_for_T_intervals.items():
        res_H_vector.append(prev_res_H + H)
        prev_res_H = res_H_vector[len(res_H_vector)-1]
    hot_util_load = - min(res_H_vector)
    assert hot_util_load>= 0
    # the lower temperature of the temperature interval for which the res_H is minimum
    pinch_cold_stream_T = all_Ts_descending[res_H_vector.index(-hot_util_load)+1]
    pinch_hot_stream_T = pinch_cold_stream_T + T_min_app
    cold_util_load = res_H_vector[len(res_H_vector)-1] + hot_util_load
    assert cold_util_load>=0
    pinch_T_arr = []
    
    for i in range(len(T_in_arr)):
        if not is_cold_stream_index(i):
            if T_in_arr[i]<pinch_hot_stream_T:
                pinch_T_arr.append(T_in_arr[i])
            elif T_out_arr[i]>pinch_hot_stream_T:
                pinch_T_arr.append(T_out_arr[i])
            else:
                pinch_T_arr.append(pinch_hot_stream_T)
        else:
            if T_in_arr[i]>pinch_cold_stream_T:
                pinch_T_arr.append(T_in_arr[i])
            elif T_out_arr[i]<pinch_cold_stream_T:
                pinch_T_arr.append(T_out_arr[i])
            else:
                pinch_T_arr.append(pinch_cold_stream_T)
    pinch_T_arr = np.array(pinch_T_arr)
    
    return pinch_T_arr, hot_util_load, cold_util_load, T_in_arr, T_out_arr,\
           T_hot_side_arr, T_cold_side_arr, hus_heating,\
           hus_cooling, hxs_heating, hxs_cooling, hxs, hot_indices,\
           cold_indices, streams, hx_utils_rearranged
            
        
def load_duties(streams, pinch_T_arr, T_out_arr, indices, is_cold, Q_hot_side, Q_cold_side):
    for index in indices:
        stream = streams[index].copy()
        H_in = stream.H
        stream.T = pinch_T_arr[index]
        stream.vle(T = pinch_T_arr[index], P = stream.P)
        H_pinch = stream.H
        stream.T = T_out_arr[index]
        stream.vle(T = T_out_arr[index], P = stream.P)
        H_out = stream.H
        if not is_cold(index):
            dH1 = abs(H_pinch - H_in)
            dH2 = abs(H_out - H_pinch)
            if abs(dH1)<0.01: dH1 = 0
            if abs(dH2)<0.01: dH2 = 0
            Q_hot_side[index] = ['cool', dH1]
            Q_cold_side[index] = ['cool', dH2]
        else:
            dH1 = H_out - H_pinch
            dH2 = H_pinch - H_in
            if abs(dH1)<0.01: dH1 = 0
            if abs(dH2)<0.01: dH2 = 0
            Q_hot_side[index] = ['heat', dH1]
            Q_cold_side[index] = ['heat', dH2]
            
            
def get_T_transient(pinch_T_arr, indices, T_in_arr):
    T_transient = pinch_T_arr.copy()
    T_transient[indices] = T_in_arr[indices]
    return T_transient


def synthesize_network(hus, ID_original, T_min_app=5.):  
    bst.main_flowsheet.set_flowsheet('%s_HXN'%ID_original)
    pinch_T_arr, hot_util_load, cold_util_load, T_in_arr, T_out_arr,\
        T_hot_side_arr, T_cold_side_arr, hus_heating, hus_cooling, hxs_heating,\
        hxs_cooling, hxs, hot_indices, cold_indices, streams, hx_utils_rearranged =\
        temperature_interval_pinch_analysis(hus, T_min_app = T_min_app, ID_original = None)        
    duties = np.array([abs(hx.Q) for hx in hxs])
    C_flow_vector = duties/np.abs(T_in_arr - T_out_arr)
    Q_hot_side = {}
    Q_cold_side = {}
    T_transient_hot_side = get_T_transient(pinch_T_arr, hot_indices, T_in_arr)
    T_transient_cold_side = get_T_transient(pinch_T_arr, cold_indices, T_in_arr)
    indices = hot_indices + cold_indices
    stream_HXs_dict = {i:[] for i in indices}
    is_cold = lambda x: x in cold_indices
    load_duties(streams, pinch_T_arr, T_out_arr, indices, is_cold, Q_hot_side, Q_cold_side)
    
    matches_hs = {i: [] for i in cold_indices}
    matches_cs = {i: [] for i in hot_indices}
    candidate_hot_streams = list(hot_indices)
    candidate_cold_streams = list(cold_indices)
    HXs_hot_side = []
    HXs_cold_side = []
    
                            # ------------- Cold side design ------------- # 
    unavailables = []
    unavailables = set([i for i in hot_indices if T_out_arr[i] >= pinch_T_arr[i]])
    unavailables.update([i for i in cold_indices if T_in_arr[i] >= pinch_T_arr[i]])
    for hot in hot_indices:
        stream_quenched = False
        original_hot_stream = streams[hot]
        # T_transient_cold_side[hot]  = min(T_transient_cold_side[hot], T_transient_hot_side[hot] )
        potential_matches = []
        for cold in cold_indices:
            if (C_flow_vector[hot]>= C_flow_vector[cold] and
                    T_transient_cold_side[hot] > T_transient_cold_side[cold] + T_min_app and
                    (hot not in unavailables) and (cold not in unavailables) and
                    (cold not in matches_cs[hot]) and (cold in candidate_cold_streams)): 
                potential_matches.append(cold)
        
        potential_matches = sorted(potential_matches, key = lambda pot_cold:
                      (min(C_flow_vector[hot], C_flow_vector[pot_cold])
                       * (T_transient_cold_side[hot] 
                          - T_transient_cold_side[cold] - T_min_app)),
                      reverse = True)
        for cold in potential_matches:
            original_cold_stream = streams[cold]
            try:
                Q_hstr = Q_cold_side[hot][1]
                Q_cstr = Q_cold_side[cold][1]
                Q_res = Q_cstr - Q_hstr
                hot_stream = original_hot_stream.copy()
                hot_stream.vle(T = T_transient_cold_side[hot], P = hot_stream.P)
                cold_stream = original_cold_stream.copy()
                cold_stream.vle(T = T_transient_cold_side[cold], P = cold_stream.P)
                if abs(T_transient_cold_side[cold] - pinch_T_arr[cold])<= 0.01:
                    continue
                ID = 'HX_%s_%s_cs'%(hot, cold)
                hot_stream.ID = 's_%s__%s'%(hot,ID)
                cold_stream.ID = 's_%s__%s'%(cold,ID)
                outsIDs = ('%s__s_%s'%(ID,hot), '%s__s_%s'%(ID,cold))
                new_HX = bst.units.HXprocess(ID = ID, ins = (hot_stream, cold_stream),
                         outs = outsIDs, T_lim0 = T_out_arr[hot],
                         T_lim1 = pinch_T_arr[cold], dT = T_min_app)
                new_HX.simulate()
                HXs_cold_side.append(new_HX)
                stream_HXs_dict[hot].append(new_HX)
                stream_HXs_dict[cold].append(new_HX)
                Q_cold_side[hot][1] -= new_HX.Q
                Q_cold_side[cold][1] -= new_HX.Q
                T_transient_cold_side[hot] = new_HX.outs[0].T
                T_transient_cold_side[cold] = new_HX.outs[1].T
                stream_quenched = T_transient_cold_side[hot] <= T_out_arr[hot]
                matches_cs[hot].append(cold)
            except:
                pass
            if stream_quenched:
                break
                
                            # ------------- Hot side design ------------- #                            
    unavailables = set([i for i in hot_indices if T_in_arr[i] <= pinch_T_arr[i]])
    unavailables.update([i for i in cold_indices if T_out_arr[i] <= pinch_T_arr[i]])
    for cold in cold_indices:
        stream_quenched = False
        original_cold_stream = streams[cold]
        T_transient_hot_side[cold] = min(T_transient_hot_side[cold], T_transient_cold_side[cold])
        for hot in candidate_hot_streams:
            if ((cold in matches_cs.keys() and hot in matches_cs[cold])
                or (cold in matches_hs.keys() and hot in matches_hs[cold])):
                pass
            potential_matches = []
            if (C_flow_vector[cold]>= C_flow_vector[hot] and
                    T_transient_hot_side[hot] > T_transient_hot_side[cold] + T_min_app and
                    (hot not in unavailables) and (cold not in unavailables) and
                    (hot not in matches_hs[cold]) and (hot in candidate_hot_streams)):
                potential_matches.append(hot)
                
        potential_matches = sorted(potential_matches, key = lambda pot_hot:
                                   (min(C_flow_vector[cold], C_flow_vector[pot_hot])
                                       * ( T_transient_hot_side[pot_hot] 
                                          - T_transient_hot_side[cold] - T_min_app)),
                                   reverse = True)
            
        for hot in potential_matches:
            original_hot_stream = streams[hot]
            try:
                Q_hstr = Q_hot_side[hot][1]
                Q_cstr = Q_hot_side[cold][1]
                Q_res = Q_cstr - Q_hstr
                cold_stream = original_cold_stream.copy()
                cold_stream.vle(T = T_transient_hot_side[cold], P = cold_stream.P)
                hot_stream = original_hot_stream.copy()
                hot_stream.vle(T = T_transient_hot_side[hot], P = hot_stream.P)
                if abs(T_transient_hot_side[hot] - pinch_T_arr[hot])<= 0.01:
                    continue
                ID = 'HX_%s_%s_hs'%(cold, hot)
                cold_stream.ID = 's_%s__%s'%(cold,ID)
                hot_stream.ID = 's_%s__%s'%(hot,ID)
                outsIDs = ('%s__s_%s'%(ID,cold), '%s__s_%s'%(ID,hot))
                new_HX = bst.units.HXprocess(ID = ID, ins = (cold_stream, hot_stream),
                         outs = outsIDs, T_lim0 = T_out_arr[cold],
                         T_lim1 = pinch_T_arr[hot], dT = T_min_app)
                new_HX.simulate()
                HXs_hot_side.append(new_HX)
                stream_HXs_dict[hot].append(new_HX)
                stream_HXs_dict[cold].append(new_HX)
                Q_hot_side[hot][1] -= new_HX.Q
                Q_hot_side[cold][1] -= new_HX.Q
                T_transient_hot_side[cold] = new_HX.outs[0].T
                T_transient_hot_side[hot] = new_HX.outs[1].T
                stream_quenched = T_transient_hot_side[cold] >= T_out_arr[cold]
                matches_hs[cold].append(hot)
            except:
                pass
            if stream_quenched:
                break
                
    # Offset heating requirement on cold side
    for cold in cold_indices:
        original_cold_stream = streams[cold]
        if Q_cold_side[cold][0]=='heat' and Q_cold_side[cold][1]>0:
            for hot in hot_indices:
                if ((cold in matches_cs.keys() and hot in matches_cs[cold])
                    or (cold in matches_hs.keys() and hot in matches_hs[cold])):
                    break
                original_hot_stream = streams[hot]
                if (Q_cold_side[hot][0]=='cool' and Q_cold_side[hot][1]>0 and
                        T_transient_cold_side[hot] - T_transient_cold_side[cold] >= T_min_app):
                    try:
                        hot_stream = original_hot_stream.copy()
                        hot_stream.vle(T = T_transient_cold_side[hot], P = hot_stream.P)
                        cold_stream = original_cold_stream.copy()
                        cold_stream.vle(T = T_transient_cold_side[cold], P = cold_stream.P)
                        if abs(T_transient_cold_side[cold] - pinch_T_arr[cold])<= 0.01:
                            continue
                        ID = 'HX_%s_%s_cs'%(hot, cold)
                        hot_stream.ID = 's_%s__%s'%(hot,ID)
                        cold_stream.ID = 's_%s__%s'%(cold,ID)
                        outsIDs = ('%s__s_%s'%(ID,hot), '%s__s_%s'%(ID,cold))
                        new_HX = bst.units.HXprocess(ID = ID, ins = (hot_stream, cold_stream),
                                 outs = outsIDs, T_lim0 = T_out_arr[hot],
                                 T_lim1 = pinch_T_arr[cold], dT = T_min_app)
                        new_HX.simulate()
                        HXs_cold_side.append(new_HX)
                        stream_HXs_dict[hot].append(new_HX)
                        stream_HXs_dict[cold].append(new_HX)                    
                        Q_cold_side[hot][1] -= new_HX.Q
                        Q_cold_side[cold][1] -= new_HX.Q                        
                        T_transient_cold_side[hot] = new_HX.outs[0].T
                        T_transient_cold_side[cold] = new_HX.outs[1].T                        
                        matches_cs[hot].append(cold)
                    except:
                        pass     
                    
    # Offset cooling requirement on hot side    
    for hot in hot_indices:
        original_hot_stream = streams[hot]
        if Q_hot_side[hot][0]=='cool' and Q_hot_side[hot][1]>0:
            for cold in cold_indices:
                if ((hot in matches_cs.keys() and cold in matches_cs[hot])
                    or (hot in matches_hs.keys() and cold in matches_hs[hot])):
                    break
                original_cold_stream = streams[cold]
                if (Q_hot_side[cold][0]=='heat' and Q_hot_side[cold][1]>0 and
                        T_transient_hot_side[hot] - T_transient_hot_side[cold] >= T_min_app):                    
                    try:
                        cold_stream = original_cold_stream.copy()
                        cold_stream.vle(T = T_transient_hot_side[cold], P = cold_stream.P)                        
                        hot_stream = original_hot_stream.copy()
                        hot_stream.vle(T = T_transient_hot_side[hot], P = hot_stream.P)                        
                        if abs(T_transient_hot_side[hot] - pinch_T_arr[hot])<= 0.01:
                            continue                        
                        ID = 'HX_%s_%s_hs'%(cold, hot)
                        cold_stream.ID = 's_%s__%s'%(cold,ID)
                        hot_stream.ID = 's_%s__%s'%(hot,ID)
                        outsIDs = ('%s__s_%s'%(ID,cold), '%s__s_%s'%(ID,hot))
                        new_HX = bst.units.HXprocess(ID = ID, ins = (cold_stream, hot_stream),
                                 outs = outsIDs, T_lim0 = T_out_arr[cold], 
                                 T_lim1 = pinch_T_arr[hot], dT = T_min_app)
                        new_HX.simulate()
                        HXs_hot_side.append(new_HX)                        
                        stream_HXs_dict[hot].append(new_HX)
                        stream_HXs_dict[cold].append(new_HX)                        
                        Q_hot_side[hot][1] -= new_HX.Q
                        Q_hot_side[cold][1] -= new_HX.Q                        
                        T_transient_hot_side[cold] = new_HX.outs[0].T
                        T_transient_hot_side[hot] = new_HX.outs[1].T       
                        stream_quenched = T_transient_hot_side[cold] >= T_out_arr[cold]                    
                        matches_hs[cold].append(hot)                    
                    except:
                        pass     
                    
    # Add final utility HXs
    new_HX_utils = []    
    for hot in hot_indices:
        new_HX_util = None
        if T_transient_cold_side[hot] > T_out_arr[hot]:
            original_hot_stream = streams[hot]
            hot_stream = original_hot_stream.copy()
            hot_stream.vle(T = T_transient_cold_side[hot], P = hot_stream.P)
            ID = 'Util_%s_cs'%(hot)
            hot_stream.ID = 's_%s__%s'%(hot,ID)
            outsID = '%s__s_%s'%(ID,hot)
            new_HX_util = bst.units.HXutility(ID = ID, ins = hot_stream,outs = outsID,
                                              T = T_out_arr[hot], rigorous = True)
            new_HX_util.simulate()
            new_HX_utils.append(new_HX_util)
            stream_HXs_dict[hot].append(new_HX_util)
        if T_transient_hot_side[hot] > pinch_T_arr[hot] + 0.05:
            original_hot_stream = streams[hot]
            hot_stream = original_hot_stream.copy()
            hot_stream.vle(T = T_transient_hot_side[hot], P = hot_stream.P)
            ID = 'Util_%s_hs'%(hot)
            hot_stream.ID = 's_%s__%s'%(hot,ID)
            outsID = '%s__s_%s'%(ID,hot)
            new_HX_util = bst.units.HXutility(ID = ID, ins = hot_stream, outs = outsID,
                                              T = pinch_T_arr[hot], rigorous = True)
            new_HX_util.simulate()
            new_HX_utils.append(new_HX_util)
            stream_HXs_dict[hot].append(new_HX_util)
    for cold in cold_indices:
        if T_transient_hot_side[cold] < T_out_arr[cold]:
            original_cold_stream = streams[cold]
            cold_stream = original_cold_stream.copy()
            cold_stream.vle(T = T_transient_hot_side[cold], P = cold_stream.P)
            ID = 'Util_%s_hs'%(cold)
            cold_stream.ID = 's_%s__%s'%(cold,ID)
            outsID = '%s__s_%s'%(ID,cold)
            new_HX_util = bst.units.HXutility(ID = ID, ins = cold_stream, outs = outsID,
                                              T = T_out_arr[cold], rigorous = True)
            new_HX_util.simulate()
            new_HX_utils.append(new_HX_util)
            stream_HXs_dict[cold].append(new_HX_util)
        # if T_transient_cold_side[cold] + 0.05 < pinch_T_arr[cold]:
        #     original_cold_stream = streams[cold]
        #     cold_stream = original_cold_stream.copy()
        #     cold_stream.vle(T = T_transient_cold_side[cold], P = cold_stream.P)
        #     ID = 'Util_%s_cs'%(cold)
        #     cold_stream.ID = 's_%s__%s'%(cold,ID)
        #     outsID = '%s__s_%s'%(ID,cold)
        #     new_HX_util = bst.units.HXutility(ID = ID, ins = cold_stream, outs = outsID,
        #                                       T = pinch_T_arr[cold], rigorous = True)
        #     new_HX_util.simulate()
        #     new_HX_utils.append(new_HX_util)
        #     stream_HXs_dict[cold].append(new_HX_util)
            
    new_hus = bst.process_tools.heat_exchanger_utilities_from_units(new_HX_utils)
    act_cool_util_load = sum([abs(hu.duty) for hu in new_hus if hu.duty<0])
    act_heat_util_load = sum([hu.duty for hu in new_hus if hu.duty>0])
    orig_heat_util_load = sum([hu.duty for hu in hus_heating])
    orig_cool_util_load = sum([abs(hu.duty) for hu in hus_cooling])
    Q_prev_heating = sum([hx.Q for hx in hxs_heating])
    Q_prev_cooling = sum([abs(hx.Q) for hx in hxs_cooling])
    Q_prev = Q_prev_heating + Q_prev_cooling
    Q_HXp = sum([hx.Q for hx in HXs_hot_side]) + sum([hx.Q for hx in HXs_cold_side])
    Q_new_heating = sum([hx_util.Q for hx_util in new_HX_utils
                         if hx_util.ins[0].T < hx_util.outs[0].T])
    Q_new_cooling = sum([abs(hx_util.Q) for hx_util in new_HX_utils
                         if hx_util.ins[0].T > hx_util.outs[0].T])
    Q_new_utils = Q_new_heating + Q_new_cooling
    Q_new = 2*Q_HXp + Q_new_utils
    Q_bal = Q_new/Q_prev
    Q_percent_error = 100*(Q_bal - 1)
    if abs(Q_percent_error)>2:
        msg = '\n\n\n WARNING: Q balance of HXN off by %s p.c.,\ which is more than 2 p.c.\n\n\n'\
              %(Q_percent_error)
        warn(msg, UserWarning, stacklevel =2)
    
    return matches_hs, matches_cs, Q_hot_side, Q_cold_side, unavailables, act_heat_util_load,\
           act_cool_util_load, HXs_hot_side, HXs_cold_side, new_HX_utils, hxs, T_in_arr,\
           T_out_arr, pinch_T_arr, C_flow_vector, hx_utils_rearranged, streams, stream_HXs_dict,\
           hot_indices, cold_indices, orig_heat_util_load, orig_cool_util_load, Q_percent_error



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
General functional algorithms for batch design.

"""

__all__ = ('size_batch',)

def size_batch(F_vol, tau_reaction, tau_cleaning, N_reactors, V_wf) -> dict:
    r"""
    Solve for batch reactor volume, cycle time, and loading time.
    
    Parameters
    ----------
    F_vol : float
        Volumetric flow rate.
    tau_reaction : float
        Reaction time.
    tau_cleaning : float
        Cleaning in place time.
    N_reactors : int
        Number of reactors.
    V_wf : float
        Fraction of working volume.
    
    Returns
    -------
    dict
        * 'Reactor volume': float
        * 'Batch time': float
        * 'Loading time': float
       
    Notes
    -----
    By assuming no downtime, the total volume of all reactors is:
        
    .. math::
        V_T = F_{vol}(\tau_{reaction} + \tau_{cleaning} + \tau_{loading})
        
    where :math:`V_T` is the total volume of all reactors, :math:`F_{vol}` is the 
    volumetric flow rate of the feed, :math:`\tau_{reaction}` is the 
    reaction time, :math:`\tau_{cleaning}` is the cleaning and unloading time, 
    and :math:`\tau_{loading}` is the time required to load a vessel. This
    equation makes the conservative assumption that no reaction takes place 
    when the tank is being filled.
    
    The working volume of an individual reactor is:
    
    .. math::
    
        V_{i,working} = \frac{V_T}{N_{reactors}}
    
    where :math:`N_{reactors}` is the number of reactor vessels.
    
    The time required to load a reactor (assuming no downtime) is:
        
    .. math::

        \tau_{loading} = \frac{V_{i,working}}{F_{vol}}
    
    Note that the the actual volume of a reaction is:
        
    .. math::

        V_i = \frac{V_{i,working}}{f}
        
    where f is the fraction of working volume in a reaction.
    
    Plugging in and solving for the total volume, :math:`V_{T}`, we have: 
    
    .. math::

        V_T = F_{vol}\frac{\tau_{reaction} + \tau_{cleaning}}{1 - \frac{1}{N_{reactors}}}
    
    Using this equation, :math:`V_T` is first calculated, then :math:`V_{i, working}`, 
    :math:`\tau_{loading}`, and :math:`V_i`.
    
    Units of measure may vary so long as they are consistent. The loading time
    can be considered the cycle time in this scenario.
    
    """
    # Total volume of all reactors, assuming no downtime
    V_T = F_vol * (tau_reaction + tau_cleaning) / (1 - 1 / N_reactors)
    
    # Volume of an individual reactor
    V_i = V_T/N_reactors
    
    # Time required to load a reactor
    tau_loading = V_i/F_vol
    
    # Total batch time
    tau_batch = tau_reaction + tau_cleaning + tau_loading
    
    # Account for excess volume
    V_i /= V_wf 
    
    return {'Reactor volume': V_i,
            'Batch time': tau_batch,
            'Loading time': tau_loading}
        
        


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .. import Unit
from thermosteam import Stream
from inspect import signature
from ._parameter import Parameter

__all__ = ('Block',)

def do_nothing(): pass

# %% Simulation blocks

class Block:
    """
    Create a Block object that can simulate the element and the system 
    downstream. The block can also generate Parameter objects that can
    update the system state.
    
    Parameters
    ----------
    element : Unit or :class:`~thermosteam.Stream`
        Element in the system.  
    system=None : System, optional
        System affected by element. In other words, element should affect
        downstream operations in the system.
        
    """
    
    _blocks = {}
    __slots__ = ('_system', '_simulate', '_element')
    
    def __new__(cls, element, system=None):
        block = cls._blocks.get((id(system), element))
        if block:
            self = block
        else:
            self = object.__new__(cls)
            self.__init__(element, system)
        return self
    
    def __init__(self, element, system=None):
        isa = isinstance
        if isa(element, Stream): unit = element.sink
        elif isa(element, Unit): unit = element
        else: unit = None
        if system and element:
            subsys = system._downstream_system(unit)
            simulate = subsys.simulate
        else:
            subsys = system
            simulate = unit.simulate if unit else do_nothing
        self._system = subsys
        self._simulate = simulate
        self._element = element
        self._blocks[id(system), element] = self
    
    def parameter(self, setter, simulate, name, distribution, units, 
                  baseline, bounds) -> Parameter:
        """Return a Parameter object."""
        if simulate is None: simulate = self._simulate
        if not name: name, *_ = signature(setter).parameters.keys()
        return Parameter(name, setter, simulate, self._element,
                         self._system, distribution, units, baseline, bounds)
    
    @property
    def system(self):
        """System that the block simulates."""
        return self._system
    
    @property
    def element(self):
        """Starting element of the system (if any)."""
        return self._element
    
    @property
    def simulate(self):
        """Simulate block."""
        return self._simulate
        
    def __repr__(self):
        if self._system:
            return f'<{type(self).__name__}: {self._system}>'
        elif self._element:
            if isinstance(self._element, type):
                return f"<{type(self).__name__}: {self._element.__name__}>"
            return f"<{type(self).__name__}: {type(self._element).__name__}-{self._element}>"
        else:
            return f"<{type(self).__name__}>"
    
    
    
    
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""

from ._heat_exchanger_network import *
from .hxn_synthesis import *

from . import _heat_exchanger_network
from . import hxn_synthesis

__all__ = (
    *_heat_exchanger_network.__all__,
    *hxn_synthesis.__all__
)



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
import biosteam as bst
from .._heat_utility import HeatUtility

__all__ = ('heat_exchanger_utilities_from_units',
           'units_with_costs',
           'get_utility_flow',
           'get_utility_duty',
           'get_power_utilities',
           'get_heat_utilities',
           'get_installed_cost',
           'get_cooling_duty',
           'get_heating_duty',
           'get_electricity_consumption',
           'get_electricity_production',
           'get_tea_results',
           'group_by_lines',
           'group_by_types',
           'filter_by_types',
           'filter_by_lines',
           'volume_of_chemical_in_units',
           'set_construction_material',
           'set_construction_material_to_stainless_steel',
           'set_construction_material_to_carbon_steel',
           'default_utilities',
)
    
def units_with_costs(units):
    """Filter units that have a cost or design method."""
    return [i for i in units if i._cost or i._design]

def heat_exchanger_utilities_from_units(units):
    """Return a list of heat utilities from all heat exchangers,
    including the condensers and boilers of distillation columns and
    flash vessel heat exchangers."""
    heat_utilities = sum([i.heat_utilities for i in units], ())
    return [i for i in heat_utilities if i.heat_exchanger]

def group_by_lines(units): 
    """Return a dictionary of lists of units grouped by line."""
    groups = {i.line: [] for i in units}
    for i in units: groups[i.line].append(i)
    return groups

def group_by_types(units): 
    """Return a dictionary of lists of units grouped by type."""
    groups = {i.__class__: [] for i in units}
    for i in units: groups[i.__class__].append(i)
    return groups

def filter_by_types(units, types):
    """Filter units by type(s)."""
    isa = isinstance
    return [i for i in units if isa(i, types)]

def filter_by_lines(units, lines):
    """Filter units by line(s)."""
    return [i for i in units if i.line in lines]

def get_utility_flow(heat_utilities, agent):
    """Return the total utility duty of heat utilities for given agent in GJ/hr"""
    if isinstance(agent, str):
        agent = HeatUtility.get_agent(agent)
    return sum([i.flow * i.agent.MW for i in heat_utilities if i.duty > 0. and i.agent is agent]) / 1e3

def get_utility_duty(heat_utilities, agent):
    """Return the total utility duty of heat utilities for given agent in GJ/hr"""
    if isinstance(agent, str):
        agent = HeatUtility.get_agent(agent)
    return sum([i.duty for i in heat_utilities if i.duty > 0. and i.agent is agent]) / 1e6 

def get_power_utilities(units):
    """Return a list of all PowerUtility objects."""
    return [i.power_utility for i in units if i.power_utility]

def get_heat_utilities(units):
    """Return a list of all HeatUtility objects."""
    return sum([i.heat_utilities for i in units], ())

def get_installed_cost(units):
    """Return the total installed equipment cost of all units in million USD."""
    return sum([i.installed_cost for i in units]) / 1e6 # millions USD

def get_cooling_duty(heat_utilities):
    """Return the total cooling duty of all heat utilities in GJ/hr."""
    cooling_duty = sum([i.duty for i in heat_utilities if i.duty < 0 and i.flow > 0]) / 1e6 # GJ/hr
    return abs(cooling_duty)
               
def get_heating_duty(heat_utilities):
    """Return the total heating duty of all heat utilities in GJ/hr."""
    return sum([i.duty for i in heat_utilities if i.duty > 0 and i.flow > 0]) / 1e6 # GJ/hr
          
def get_electricity_consumption(power_utilities):
    """Return the total electricity consumption of all PowerUtility objects in MW."""
    return sum([i.consumption for i in power_utilities]) / 1000 # MW

def get_electricity_production(power_utilities):
    """Return the total electricity production of all PowerUtility objects in MW."""
    return sum([i.production for i in power_utilities]) / 1000 # MW

def volume_of_chemical_in_units(units, chemical):
    """Return volume of chemical that occupied in given units [m^3]."""
    isa = isinstance
    F_vol = 0.
    for i in units:
        if isa(i, bst.BatchBioreactor):
            feed = i.ins[0]
            z_vol = feed.ivol[chemical] / feed.F_vol
            D = i.design_results
            F_vol += D['Number of reactors'] * D['Reactor volume'] * z_vol
        elif isa(i, bst.Tank):
            feed = i.ins[0]
            z_vol = feed.ivol[chemical] / feed.F_vol
            D = i.design_results
            F_vol += D['Total volume'] * z_vol
    return F_vol

def set_construction_material(units, pressure_vessel_material, tray_material,
                              tank_material, heat_exchanger_material, pump_material):
    """
    Set the construction material of all vessels, columns, trays,
    heat exchangers, and pumps
    
    Parameters
    ----------
    units : Iterable[Unit]
        Unit operations to set construction material.
    pressure_vessel_material : str
        Construction material for pressure vessels (includes distillation 
        columns and flash vessels).
    tray_material : str
        Construction material for column trays.
    tank_material : str
        Construction material for tanks.
    pump_material : str
        Construction material for pumps.
    
    """
    isa = isinstance
    HeatExchanger = bst.HX
    Distillation = bst.BinaryDistillation
    PressureVessel = bst.units.design_tools.PressureVessel
    Tank = bst.Tank
    Pump = bst.Pump
    for u in units:
        for i in (u, *u.auxiliary_units):
            if isa(i, HeatExchanger):
                i.material = heat_exchanger_material
            elif isa(i, Distillation):
                i.vessel_material = pressure_vessel_material
                i.tray_material = tray_material
            elif isa(i, PressureVessel):
                i.vessel_material = pressure_vessel_material
            elif isa(i, Tank):
                i.vessel_material = tank_material
            elif isa(i, Pump):
                i.material = pump_material

def set_construction_material_to_stainless_steel(units, kind='304'):
    """
    Set the construction material of all vessels, columns, trays,
    heat exchangers, and pumps to stainless steel.
    
    Parameters
    ----------
    kind : str, "304" or "316"
        Type of stainless steel. 
    
    """
    if kind not in ('304', '316'): 
        raise ValueError("kind must be either '304' or '316', not '%s'" %kind)
    material = 'Stainless steel ' + kind
    set_construction_material(units, material, material, 'Stainless steel',
                              'Stainless steel/stainless steel', 
                              'Stainless steel')

def set_construction_material_to_carbon_steel(units):
    """
    Set the construction material of all vessels, columns, trays,
    heat exchangers, and pumps to carbon steel or cast iron.
    
    """
    set_construction_material(units, 'Carbon steel', 'Carbon steel', 
                              'Carbon steel', 'Carbon steel/carbon steel', 
                              'Cast iron')

def get_tea_results(tea, product=None, feedstock=None):
    """Return a dictionary with general TEA results."""
    TCI = tea.TCI / 1e6
    VOC = tea.VOC / 1e6
    FOC = tea.FOC / 1e6
    installed_cost = tea.installed_cost / 1e6
    material_cost = tea.material_cost / 1e6
    utility_cost = tea.utility_cost / 1e6
    sales = tea.sales / 1e6 
    dct = {
        'TCI [million USD]': round(TCI),
        'Installed equipment cost [million USD]': round(installed_cost),
        'VOC [million USD/yr]': round(VOC),
        'FOC [million USD/yr]': round(FOC),
        'Material cost [million USD/yr]': round(material_cost),
        'Utility cost [million USD/yr]': round(utility_cost),
        'Sales [million USD/yr]': round(sales),
    }
    if product:
        MPSP = tea.solve_price(product) * 907.18474
        dct['MPSP [USD/ton]'] = round(MPSP)
    if feedstock:
        MFP = tea.solve_price(feedstock) * 907.18474
        dct['MFP [USD/ton]'] = round(MFP)
    return dct
        
def default_utilities():
    """Reset utilities back to BioSTEAM's defaults."""
    bst.HeatUtility.default_agents()
    bst.PowerUtility.default_price()
    


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from scipy.optimize import brentq

__all__ = ('BoundedNumericalSpecification',)

class BoundedNumericalSpecification: # pragma: no cover
    __slots__ = ('args', 'solver', 'kwargs')
    
    def __init__(self, *args, solver=brentq, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.solver = solver
        
    def __call__(self):
        return self.solver(*self.args, **self.kwargs)
    